from typing import Any, Literal

import jinja2
from pydantic import BaseModel

from .completion import FunctionMessage, Message


class MessageTemplate(BaseModel):
    role: Literal["system", "user", "assistant", "function"]
    content: str
    name: str = ""

    def render(self, **kwargs: Any) -> Message:
        content = jinja2.Template(self.content).render(**kwargs)
        if self.role == "function":
            return FunctionMessage(
                role=self.role,
                content=content,
                name=self.name,
            )
        return Message(role=self.role, content=content)

    class Config:
        # remove whitespace
        anystr_strip_whitespace = True


class ChatTemplate(BaseModel):
    messages: list[MessageTemplate]

    def render(self, **kwargs: Any) -> list[Message]:
        return [m.render(**kwargs) for m in self.messages]



import os
from typing import Any

import pydantic
import pytest
from syrupy.assertion import SnapshotAssertion

from ..completion import APISettings, achat_completion, afunction_completion, astructural_completion, chat_completion, function_completion, structural_completion
from ..exceptions import CompletionIncompleteError


def exclude_api_settings(result: dict[str, Any]) -> dict[str, Any]:
    request: dict[str, Any] = result["request"]
    request.pop("api_base", None)
    request.pop("api_key", None)
    request.pop("api_type", None)
    request.pop("api_version", None)

    return result


def book_a_flight(date: str, destination: str, origin: str = "London") -> str:
    return f"Booked a flight from {origin} to {destination} on {date}."


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
def test_function_completion(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    msg = function_completion(
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "I want to book a flight to London."},
        ],
        functions=[book_a_flight],
        api_settings=api_settings,
        model=model,
    )
    assert snapshot == msg


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
async def test_afunction_completion(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    msg = await afunction_completion(
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "I want to book a flight to London."},
        ],
        functions=[book_a_flight],
        api_settings=api_settings,
        model=model,
    )
    assert snapshot == msg


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
def test_function_completion_without_enough_tokens(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    with pytest.raises(CompletionIncompleteError) as excinfo:
        msg = function_completion(
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "I want to book a restaurant in London."},
            ],
            max_tokens=10,
            functions=[book_a_flight],
            api_settings=api_settings,
            model=model,
        )
    assert snapshot == excinfo.exconly()
    assert snapshot == exclude_api_settings(vars(excinfo.value))


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
def test_function_completion_without_neccess_function(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    with pytest.raises(CompletionIncompleteError) as excinfo:
        msg = function_completion(
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "I want to book a restaurant in London."},
            ],
            functions=[book_a_flight],
            api_settings=api_settings,
            model=model,
        )
    assert snapshot == excinfo.exconly()
    assert snapshot == exclude_api_settings(vars(excinfo.value))


class Email(pydantic.BaseModel):
    subject: str = pydantic.Field(description="the subject of email")
    body: str = pydantic.Field(description="the body of email")


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
def test_structural_completion(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    email = structural_completion(
        Email,
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Give me a sample email."},
        ],
        api_settings=api_settings,
        model=model,
    )

    assert snapshot == email


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
async def test_astructural_completion(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    email = await astructural_completion(
        Email,
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Give me a sample email."},
        ],
        api_settings=api_settings,
        model=model,
    )

    assert snapshot == email


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
def test_structural_completion_without_enough_tokens(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    with pytest.raises(CompletionIncompleteError) as excinfo:
        email = structural_completion(
            Email,
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Give me a sample email."},
            ],
            max_tokens=10,
            api_settings=api_settings,
            model=model,
        )
    assert snapshot == excinfo.exconly()
    assert snapshot == exclude_api_settings(vars(excinfo.value))


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
def test_chat_completion(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    msg = chat_completion(
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Hello, who are you?"},
        ],
        api_settings=api_settings,
        model=model,
    )
    assert snapshot == msg


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
async def test_achat_completion(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    msg = await achat_completion(
        [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Hello, who are you?"},
        ],
        api_settings=api_settings,
        model=model,
    )
    assert snapshot == msg


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize(
    "api_settings, model",
    [
        pytest.param(
            APISettings(
                api_key=str(os.environ.get("OPENAI_API_KEY")),
                api_type="open_ai",
            ),
            "gpt-3.5-turbo-0613",
            id="OpenAI API",
        ),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-07-01-preview"), "gpt35", id="Azure 2023-07-01 API"),
        pytest.param(APISettings(api_key=str(os.environ.get("AZURE_OPENAI_API_KEY")), api_base="https://gstudio.openai.azure.com", api_type="azure", api_version="2023-12-01-preview"), "gpt35", id="Azure 2023-12-01 API"),
    ],
)
def test_chat_completion_incomplete(snapshot: SnapshotAssertion, api_settings: APISettings, model: str) -> None:
    with pytest.raises(CompletionIncompleteError) as excinfo:
        chat_completion(
            [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Hello, who are you?"},
            ],
            max_tokens=1,
            api_settings=api_settings,
            model=model,
        )

    assert snapshot == excinfo.exconly()
    assert snapshot == exclude_api_settings(vars(excinfo.value))



from typing import Any

from openai.types.chat import ChatCompletion


class GptFnError(Exception):
    pass


class CompletionIncompleteError(GptFnError):
    def __init__(self, msg: str, request: dict[str, Any], response: ChatCompletion) -> None:
        super().__init__(msg)
        self.response = response
        self.request = request


class AiFnError(GptFnError):
    def __init__(self, msg: str, fn_locals: dict[str, Any]) -> None:
        super().__init__(msg)
        self.fn_locals = fn_locals



from typing import Annotated, Any, Callable, Literal

import pytest
from pydantic import BaseModel, Field
from syrupy.assertion import SnapshotAssertion

from ..signature import FunctionSignature


def add(a: int, b: int = 10) -> int:  # type: ignore[empty-body]
    """Add two numbers."""


def concat(a: str, b: str) -> str:  # type: ignore[empty-body]
    """Concat two strings
    :param a: first string
    :param b: second string
    """


class Person(BaseModel):
    """Person model."""

    name: str
    age: int


def fake_person(count: int) -> Person:  # type: ignore[empty-body]
    """generate fake person."""


def no_return_annoation(a: int, b: int):  # type: ignore[no-untyped-def]
    """Add two numbers."""


def concats(*args: str) -> str:  # type: ignore[empty-body]
    """Concat the given strings"""


def complex(a: str, b: str, *args: str, c: str, d: str, **kwargs: str) -> str:  # type: ignore[empty-body]
    """Complex function"""


def get_current_weather(*locations: str, unit: Literal["celsius", "fahrenheit"]) -> list[dict[str, Any]]:
    """ "Get the current weather in a given location"""
    return [
        {
            "location": location,
            "temperature": "72",
            "forecast": ["sunny", "windy"],
        }
        for location in locations
    ]


def is_male(person: Person) -> bool:  # type: ignore[empty-body]
    """return the person is male"""


def how_many(num: Annotated[int, Field(gt=10, description="greater than 10")]) -> int:
    """return the given number"""
    return num


def add_ext(filename: str, ext: str) -> str:
    """Add the given extension to the filename

    This is a multiline docstring.

    :param filename: the filename
    :param ext: the extension, e.g. `.txt`
    """
    return filename + ext


@pytest.mark.parametrize(
    "func, args, kwargs",
    [
        (add, (1, 2), {}),
        (fake_person, (5,), {}),
        (concat, ("pen", "apple"), {}),
        (concat, ("a'b", "cd"), {}),
        (concats, ("a", "b", "c", "d"), {}),
        (
            complex,
            ("a-v", "b-v", "arg-v1", "arg-v2"),
            {"c": "c-v", "d": "d-v", "kwarg1": "kwarg1-v", "kwarg2": "kwarg2-v"},
        ),
        (get_current_weather, ("New York", "London"), {"unit": "celsius"}),
        (is_male, (Person(name="John", age=20),), {}),
        (how_many, (20,), {}),
        (add_ext, ("test", ".txt"), {}),
    ],
)
def test_function_signature(
    func: Callable[[Any], Any],
    args: tuple[Any],
    kwargs: dict[str, Any],
    snapshot: SnapshotAssertion,
) -> None:
    sig = FunctionSignature(func)
    assert snapshot == sig.instruction()
    assert snapshot == sig.call_line(*args, **kwargs)
    assert snapshot == sig.schema()


def test_function_signature_without_return_annoations() -> None:
    with pytest.raises(AssertionError):
        FunctionSignature(no_return_annoation)



from functools import wraps
from typing import Any, Callable, ParamSpec, TypeVar

import openai

from .completion import chat_completion
from .exceptions import AiFnError
from .prompt import ChatTemplate, MessageTemplate
from .utils.signature import FunctionSignature

T = TypeVar("T")
P = ParamSpec("P")


def ai_fn(model: str = "gpt-3.5-turbo") -> Callable[[Callable[P, T]], Callable[P, T]]:
    def _ai_fn(
        fn: Callable[P, T],
    ) -> Callable[P, T]:
        sig = FunctionSignature(fn)

        @wraps(fn)
        def inner(*args: Any, **kwargs: Any) -> T:
            fn_call = sig.call_line(*args, **kwargs)
            template = ChatTemplate(
                messages=[
                    MessageTemplate(role="system", content=sig.instruction()),
                    MessageTemplate(role="user", content=fn_call),
                ]
            )

            try:
                resp = chat_completion(template.render(), temperature=0.0, model=model)
            except openai.APIError as e:
                fn_locals = sig.locals(*args, **kwargs)
                raise AiFnError(fn_call, fn_locals) from e

            return sig.parse(resp)

        return inner

    return _ai_fn



from pathlib import Path

import pytest
from pydantic import BaseModel, Field
from syrupy.assertion import SnapshotAssertion

from ...completion import Message, structural_completion
from ..pydantic_parser import ParserError, PydanticParser


class Receipt(BaseModel):
    """The Receipt for user"""

    amount: float
    currency: str = Field(description="ISO 4217 currency code")
    customer: str


class Email(BaseModel):
    subject: str = Field(description="the subject of email")
    body: str = Field(description="the body of email")


def test_pydantic_parser_parse(snapshot: SnapshotAssertion) -> None:
    assert snapshot == PydanticParser[Email](pydantic_model=Email).parse(
        """json\n{\n  "subject": "Revolutionizing the E-Bike Industry with Innovative Aluminum Frames",\n  "body": "Dear valued client,\\n\\nWe are excited to share with you the latest news from JIN-JI International Co., Ltd. We have revolutionized the E-Bike industry with our innovative aluminum frames for SPACIOUS Industrial Co., Ltd. in Taiwan. Our commitment to innovation and customer satisfaction has positioned us as a key player in the market.\\n\\nCEO Xiao Ming, Chen, emphasizes, \\"We provide the best quality products,\\" reflecting our dedication to excellence.\\n\\nTo learn more about our revolutionary aluminum frames and how they offer increased rigidity, stability, and safety for riders, we invite you to visit our website at https://jinji.en.taiwantrade.com/. We believe that our innovative designs will support complex structures for electric bikes, ensuring a confident and secure riding experience.\\n\\nThank you for your continued support, and we look forward to bringing you more groundbreaking developments in the future.\\n\\nBest regards,\\n[Your Name]\\nJIN-JI International Co., Ltd."\n}\n"""
    )


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
@pytest.mark.parametrize("test_filename, model", [("email.txt", Email)])
def test_pydantic_parser_with_prompt(snapshot: SnapshotAssertion, test_filename: str, model: type[BaseModel], datadir: Path) -> None:

    content = (datadir / test_filename).read_text()
    instruction = content
    assert snapshot(name="instructoin") == instruction

    result = structural_completion(model, messages=[Message(role="system", content=""), Message(role="user", content=instruction)])

    assert snapshot(name="gpt response") == result


def test_pydantic_parser_get_format_instructions(
    snapshot: SnapshotAssertion,
) -> None:
    assert snapshot == PydanticParser[Receipt](pydantic_model=Receipt).get_format_instructions()


def test_parse_output(snapshot: SnapshotAssertion) -> None:
    assert snapshot == PydanticParser[Receipt](pydantic_model=Receipt).parse('{"amount": 1.0, "currency": "USD", "customer": "John Doe"}')


def test_pydantic_parser_fail() -> None:
    with pytest.raises(ParserError):
        PydanticParser[Receipt](pydantic_model=Receipt).parse('{"amount": 1.0, "currency": "USD"}')



import os
from datetime import datetime
from typing import Any, Callable, Type, TypedDict, TypeVar

import fuzzy_json
import openai
import pydantic
from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI

from .exceptions import CompletionIncompleteError
from .utils import signature

T = TypeVar("T", bound=pydantic.BaseModel)


class Message(TypedDict):
    role: str
    content: str


class FunctionMessage(Message):
    name: str


def infer_api_type() -> str:
    """
    Infers the API type based on environment variables and sets the api_type attribute accordingly.

    Raises:
        ValueError: If the API type cannot be determined.
    """
    if os.environ.get("OPENAI_API_KEY"):
        return "open_ai"
    elif os.environ.get("AZURE_OPENAI_API_KEY"):
        return "azure"
    else:
        raise ValueError("API type could not be inferred from environment variables.")


class APISettings(pydantic.BaseModel):
    api_key: str = pydantic.Field(default_factory=lambda: openai.api_key)
    api_base: str = pydantic.Field(default_factory=lambda: openai.azure_endpoint)
    api_version: str | None = pydantic.Field(default_factory=lambda: openai.api_version)
    api_type: str = pydantic.Field(default_factory=infer_api_type)

    def check_functions_required(self) -> bool:
        if self.api_version is None:
            assert self.api_type != "azure", "api_version is required for azure"
            return False
        if self.api_type == "azure" and datetime.strptime(str(self.api_version)[:10], "%Y-%m-%d") >= datetime(2023, 12, 1):
            return False
        return True


def get_sync_client(model: str, api_settings: APISettings = APISettings()) -> OpenAI:
    if api_settings.api_type == "open_ai":
        return OpenAI(api_key=api_settings.api_key)
    elif api_settings.api_type == "azure":
        return AzureOpenAI(
            azure_endpoint=api_settings.api_base,
            api_key=api_settings.api_key,
            api_version=api_settings.api_version,
            azure_deployment=model,
        )

    raise ValueError(f"Unknown api_type {api_settings.api_type}")


def get_async_client(model: str, api_settings: APISettings = APISettings()) -> AsyncOpenAI:
    if api_settings.api_type == "open_ai":
        return AsyncOpenAI(api_key=api_settings.api_key)
    elif api_settings.api_type == "azure":
        return AsyncAzureOpenAI(
            azure_endpoint=api_settings.api_base,
            api_key=api_settings.api_key,
            api_version=api_settings.api_version,
            azure_deployment=model,
        )
    raise ValueError(f"Unknown api_type {api_settings.api_type}")


"""
The currently applied Azure OpenAI API version is 2023-07-01-preview,
which is about to expire on July, 1,
and the parameters `functions` and `function_call` have been deprecated for the OpenAI API
and since the 2023-12-01-preview version for the Azure OpenAI API version.
Therefore, the Azure OpenAI API version should be upgraded to either 2023-10-01-preview or versions later than and excluding 2023-12-01-preview.
The parameters should also be updated to `tools` and `tool_choice` for OpenAI API and Azure OpenAI API with version starting from 2023-12-01-preview.
"""


def function_completion(
    messages: list[Message],
    max_tokens: int | None = None,
    model: str = "gpt-3.5-turbo-0613",
    temperature: float = 1.0,
    top_p: float = 1.0,
    frequency_penalty: float = 0.0,
    presence_penalty: float = 0.0,
    stop: list[str] = [],
    user: str = "",
    functions: list[Callable[..., Any]] = [],
    function_call: str | dict[str, Any] = "auto",
    api_settings: APISettings = APISettings(),
) -> dict[str, Any] | None:
    assert functions, "functions must be a non-empty list of functions"

    kwargs: dict[str, Any] = dict(
        messages=messages,
        model=model,
        temperature=temperature,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        user=user,
        stop=stop or None,
    )
    client: OpenAI = get_sync_client(model, api_settings)

    functions_are_required = api_settings.check_functions_required()

    kwargs_ = dict(tools=[{"type": "function", "function": signature.FunctionSignature(f).schema()} for f in functions], tool_choice={"type": "function", "function": function_call} if type(function_call) != str else function_call)

    if functions_are_required:
        kwargs_ = dict(functions=[signature.FunctionSignature(f).schema() for f in functions], function_call=function_call)

    kwargs.update(kwargs_)

    if max_tokens is not None:
        kwargs.update(max_tokens=max_tokens)

    response = client.chat.completions.create(**kwargs)

    output = response.choices[0]
    message = output.message
    finish_reason = output.finish_reason

    if functions_are_required and message.function_call is not None and finish_reason in ["stop", "function_call"]:
        return message.function_call

    elif message.tool_calls is not None and finish_reason in ["stop", "tool_calls"]:
        return message.tool_calls[0].function

    raise CompletionIncompleteError(
        f"Incomplete response. Max tokens: {max_tokens}, Finish reason: {finish_reason} Message:{message.content}",
        response=response,
        request=kwargs,
    )


async def afunction_completion(
    messages: list[Message],
    max_tokens: int | None = None,
    model: str = "gpt-3.5-turbo-0613",
    temperature: float = 1.0,
    top_p: float = 1.0,
    frequency_penalty: float = 0.0,
    presence_penalty: float = 0.0,
    stop: list[str] = [],
    user: str = "",
    functions: list[Callable[..., Any]] = [],
    function_call: str | dict[str, Any] = "auto",
    api_settings: APISettings = APISettings(),
) -> dict[str, Any] | None:
    assert functions, "functions must be a non-empty list of functions"

    kwargs: dict[str, Any] = dict(
        messages=messages,
        model=model,
        temperature=temperature,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        user=user,
        stop=stop or None,
    )
    client: AsyncOpenAI = get_async_client(model, api_settings)

    functions_are_required = api_settings.check_functions_required()

    kwargs_ = dict(tools=[{"type": "function", "function": signature.FunctionSignature(f).schema()} for f in functions], tool_choice={"type": "function", "function": function_call} if type(function_call) != str else function_call)

    if functions_are_required:
        kwargs_ = dict(functions=[signature.FunctionSignature(f).schema() for f in functions], function_call=function_call)

    kwargs.update(kwargs_)

    if max_tokens is not None:
        kwargs.update(max_tokens=max_tokens)

    response = await client.chat.completions.create(**kwargs)
    output = response.choices[0]
    message = output.message
    finish_reason = output.finish_reason

    if functions_are_required and message.function_call is not None and finish_reason in ["stop", "function_call"]:
        return message.function_call

    elif message.tool_calls is not None and finish_reason in ["stop", "tool_calls"]:
        return message.tool_calls[0].function

    raise CompletionIncompleteError(
        f"Incomplete response. Max tokens: {max_tokens}, Finish reason: {finish_reason} Message:{message.content}",
        response=response,
        request=kwargs,
    )


def structural_completion(
    structure: Type[T],
    messages: list[Message],
    max_tokens: int | None = None,
    model: str = "gpt-3.5-turbo-0613",
    temperature: float = 1.0,
    top_p: float = 1.0,
    frequency_penalty: float = 0.0,
    presence_penalty: float = 0.0,
    user: str = "",
    auto_repair: bool = True,
    api_settings: APISettings = APISettings(),
) -> T:

    kwargs: dict[str, Any] = dict(
        messages=messages,
        model=model,
        temperature=temperature,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        user=user,
    )
    client: OpenAI = get_sync_client(model, api_settings)

    functions_are_required = api_settings.check_functions_required()

    function_call = {"name": "structural_response"}
    function = function_call | {
        "description": "Response to user in a structural way.",
        "parameters": structure.schema(),
    }

    kwargs_ = dict(
        tools=[{"type": "function", "function": function}],
        tool_choice={"type": "function", "function": function_call},
    )

    if functions_are_required:
        kwargs_ = dict(
            functions=[function],
            function_call=function_call,
        )
    kwargs.update(kwargs_)

    if max_tokens is not None:
        kwargs.update(max_tokens=max_tokens)

    response = client.chat.completions.create(**kwargs)

    output = response.choices[0]
    message = output.message
    finish_reason = output.finish_reason

    if finish_reason == "stop":
        if functions_are_required and message.function_call is not None:
            args = message.function_call.arguments
        elif message.tool_calls is not None:
            args = message.tool_calls[0].function.arguments

        parsed_json = fuzzy_json.loads(args, auto_repair)
        return pydantic.parse_obj_as(structure, parsed_json)

    raise CompletionIncompleteError(
        f"Incomplete response. Max tokens: {max_tokens}, Finish reason: {finish_reason} Message:{message.content}",
        response=response,
        request=kwargs,
    )


async def astructural_completion(
    structure: Type[T],
    messages: list[Message],
    max_tokens: int | None = None,
    model: str = "gpt-3.5-turbo-0613",
    temperature: float = 1.0,
    top_p: float = 1.0,
    frequency_penalty: float = 0.0,
    presence_penalty: float = 0.0,
    user: str = "",
    auto_repair: bool = True,
    api_settings: APISettings = APISettings(),
) -> T:

    kwargs: dict[str, Any] = dict(
        messages=messages,
        model=model,
        temperature=temperature,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        user=user,
    )
    client: AsyncOpenAI = get_async_client(model, api_settings)

    functions_are_required = api_settings.check_functions_required()

    function_call = {"name": "structural_response"}
    function = function_call | {
        "description": "Response to user in a structural way.",
        "parameters": structure.schema(),
    }

    kwargs_ = dict(
        tools=[{"type": "function", "function": function}],
        tool_choice={"type": "function", "function": function_call},
    )

    if functions_are_required:
        kwargs_ = dict(
            functions=[function],
            function_call=function_call,
        )
    kwargs.update(kwargs_)

    if max_tokens is not None:
        kwargs.update(max_tokens=max_tokens)

    response = await client.chat.completions.create(**kwargs)

    output = response.choices[0]
    message = output.message
    finish_reason = output.finish_reason

    if finish_reason == "stop":
        if functions_are_required and message.function_call is not None:
            args = message.function_call.arguments
        elif message.tool_calls is not None:
            args = message.tool_calls[0].function.arguments

        parsed_json = fuzzy_json.loads(args, auto_repair)
        return pydantic.parse_obj_as(structure, parsed_json)

    raise CompletionIncompleteError(
        f"Incomplete response. Max tokens: {max_tokens}, Finish reason: {finish_reason} Message:{message.content}",
        response=response,
        request=kwargs,
    )


def chat_completion(
    messages: list[Message],
    max_tokens: int | None = None,
    model: str = "gpt-3.5-turbo",
    temperature: float = 1.0,
    top_p: float = 1.0,
    frequency_penalty: float = 0.0,
    presence_penalty: float = 0.0,
    stop: list[str] = [],
    user: str = "",
    api_settings: APISettings = APISettings(),
) -> str:

    kwargs: dict[str, Any] = dict(
        messages=messages,
        model=model,
        temperature=temperature,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        user=user,
        stop=stop or None,
    )

    client: OpenAI = get_sync_client(model, api_settings)

    if max_tokens is not None:
        kwargs.update(max_tokens=max_tokens)

    response = client.chat.completions.create(**kwargs)

    output = response.choices[0]
    output_message = output.message.content.strip()

    if output.finish_reason != "stop":
        raise CompletionIncompleteError(
            f"Incomplete response. Max tokens: {max_tokens}, Finish reason: {output.finish_reason}",
            response=response,
            request=kwargs,
        )

    return output_message


async def achat_completion(
    messages: list[Message],
    max_tokens: int | None = None,
    model: str = "gpt-3.5-turbo",
    temperature: float = 1.0,
    top_p: float = 1.0,
    frequency_penalty: float = 0.0,
    presence_penalty: float = 0.0,
    stop: list[str] = [],
    user: str = "",
    api_settings: APISettings = APISettings(),
) -> str:

    kwargs: dict[str, Any] = dict(
        messages=messages,
        model=model,
        temperature=temperature,
        top_p=top_p,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        user=user,
        stop=stop or None,
    )

    client: AsyncOpenAI = get_async_client(model, api_settings)

    if max_tokens is not None:
        kwargs.update(max_tokens=max_tokens)

    response = await client.chat.completions.create(**kwargs)

    output = response.choices[0]
    output_message = output.message.content.strip()

    if output.finish_reason != "stop":
        raise CompletionIncompleteError(
            f"Incomplete response. Max tokens: {max_tokens}, Finish reason: {output.finish_reason}",
            response=response,
            request=kwargs,
        )

    return output_message



import pytest
from pydantic import BaseModel
from syrupy.assertion import SnapshotAssertion

from ..ai_function import ai_fn
from ..exceptions import AiFnError


@ai_fn()
def fabnocci(n: int) -> int:  # type: ignore[empty-body]
    """return fabnocci number"""


class Hero(BaseModel):
    """Hero model."""

    name: str
    age: int


@ai_fn()
def fake_hero(n: int) -> list[Hero]:  # type: ignore[empty-body]
    """generate fake hero."""


@ai_fn()
def generate_hashtags(text: str) -> list[str]:  # type: ignore[empty-body]
    """generate social media hashtags from text."""


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
def test_ai_fabnocci(snapshot: SnapshotAssertion) -> None:
    assert snapshot == fabnocci(10)


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
def test_ai_fake_hero(snapshot: SnapshotAssertion) -> None:
    heros = fake_hero(5)
    assert len(heros) == 5
    assert all(isinstance(k, Hero) for k in heros)
    assert snapshot == heros


@pytest.mark.vcr(match_on=["method", "scheme", "host", "port", "path", "query", "body"])
def test_ai_fn_max_token(snapshot: SnapshotAssertion) -> None:
    with pytest.raises(AiFnError) as excinfo:
        generate_hashtags(" ".join("hello" for _ in range(5001)))

    assert snapshot(name="exconly") == excinfo.exconly()
    assert snapshot(name="context") == excinfo.value.__context__
    assert snapshot(name="vars(exc)") == vars(excinfo.value)



from inspect import cleandoc

import pydantic
from syrupy.assertion import SnapshotAssertion

from ..prompt import ChatTemplate, MessageTemplate


def test_chat_template(snapshot: SnapshotAssertion) -> None:
    template = ChatTemplate(
        messages=[
            MessageTemplate(role="system", content="hello, what's your name?"),
            MessageTemplate(role="user", content="My name is {{name}}."),
            MessageTemplate(role="system", content="hello, {{name}} nice to meet you!"),
        ]
    )

    assert snapshot == template.render(name="John")


def test_complicated_chat_template(snapshot: SnapshotAssertion) -> None:
    class Article(pydantic.BaseModel):
        title: str
        content: str

    template = ChatTemplate(
        messages=[
            MessageTemplate(
                role="user",
                content=cleandoc(
                    """
                    Summarize the following articles:

                    keywords: {{ keywords | join(", ") }}

                    {% for article in articles %}
                    # {{ article.title }}
                    {{ article.content }}
                    {% endfor %}
                    """
                ),
            ),
        ]
    )

    assert snapshot == template.render(
        articles=[
            Article(title="Article 1", content="Content 1"),
            Article(title="Article 2", content="Content 2"),
            Article(title="Article 3", content="Content 3"),
        ],
        keywords=["keyword1", "keyword2", "keyword3"],
    )
